#include "xt_int.h"

static struct kevent *changelist = NULL;
unsigned int changelist_size = 0;

static int
_xt_changelist_reallocate(unsigned int n) {
	void *p;
	int willbe;

	n <<= 1;

	if(n < changelist_size)
		return 0;

	willbe = changelist_size << 2;
	if(willbe == 0) willbe = 32;

	p = realloc(changelist, willbe * sizeof(struct kevent));
	if(p == NULL)
		return -1;

	changelist = p;
	changelist_size = willbe;

	return 0;
}

int
poll(struct pollfd *fds, unsigned int nfds, int timeout) {
	int nchanges = 0;
	int interest;
	int ret;
	int fd;
	int n;

	if(_xt_changelist_reallocate(nfds))
		return -1;


	/*
	 * First run through the descriptors - initialize
	 * everything.
	 */
	for(n = 0; n < nfds; n++) {
		fd = fds[n].fd;
		fds[n].revents = 0;

		if(fd == -1)
			continue;

		interest = fds[n].events;

		switch(_xt_is_queued(fd, interest)) {
		case 1:
			break;
		case 0:

			if((interest & POLLIN)) {
				EV_SET(&changelist[nchanges],
					fd,	/* ident */
					EVFILT_READ,
					EV_ADD,
					0,
					0,
					NULL
				);
				nchanges++;
			}
			if((interest & POLLOUT)) {
				EV_SET(&changelist[nchanges],
					fd,	/* ident */
					EVFILT_WRITE,
					EV_ADD,
					0,
					0,
					NULL
				);
				nchanges++;
			}

			break;
		case -1:
		default:
			fds[n].revents = POLLNVAL;
			continue;
		}

		/* Lock the file descriptor */
		_xt_fd_lock(fd, interest);

		/* Indicate our interest to the events in this fd */
		if((interest & POLLIN) || !(interest & POLLOUT))
			_xt_fd_wait(fd, POLLIN);
		if((interest & POLLOUT) || !(interest & POLLIN))
			_xt_fd_wait(fd, POLLOUT);
	}


	if(timeout > 0) {
		_xt_schedule_wakeup(timeout / 10);
	} else if(timeout < 0) {
		/* Sleep "forever" */
		_xt_move2ring(_xt_current_thread, &_xt_sleeping_ring);
	} /* else !timeout */

	/*
	 * Take the shot right now if we need to change events
	 * or requested to take one shot explicitly
	 */
	if(changelist || !timeout)
		_xt_gather_events(changelist, nchanges, 0);

	if(changelist && timeout) {
		if(_xt_is_in_ring(_xt_current_thread, _xt_runnable_ring)) {
			/*
			 * So, here is the story.
			 * When _xt_gather_events() got some events,
			 * it will wake up the waiting thread, and that
			 * would be me. At this point, current thread
			 * will again be in the running ring, so we need
			 * to stop the timer.
			 */
			if(timeout > 0)
				_xt_timer_stop(&_xt_current_thread->timer);
		} else {
			/*
			 * Well, there were no events for us in that shot,
			 * so lay back and take some rest
			 */
			THR_SWITCH_CONTEXT;
		}
	}

	/*
	 * The second run though the descriptors - gather
	 * the resulting information.
	 */
	ret = 0;
	for(n = 0; n < nfds; n++) {
		if(fds[n].fd == -1) {
			fds[n].revents = POLLNVAL;
			continue;
		}

		fds[n].revents =
			_xt_fd_state(fds[n].fd)
			  & (fds[n].events | POLLNVAL | POLLHUP);

		if((fds[n].revents & (POLLIN | POLLOUT | POLLHUP)))
			ret++;

		/* Indicate the end of our inerest in this fd */
		_xt_fd_wait(fds[n].fd, 0);

		/* Unlock the file descriptor */
		_xt_fd_unlock(fds[n].fd, fds[n].events);
	}
	

	return ret;
}


