#include "xt_int.h"

static xt_fd_t fd_wrappers_static[1024];

static int xfd_size = 1024;
static xt_fd_t *xfd_wrapper = fd_wrappers_static;

int _xt_kqueue = -1;
static struct kevent keventlist_static[128];

/* Convert file descriptor to the internal representation */
static xt_fd_t *_xt_get_xfd(int fd);

/* Reinitialize the wrappers table */
static int _xt_xfd_reinit();


int
_xt_fd_nbio(int fd, int socket_way) {
	if(fd == -1)
		return -1;

	if(socket_way) {
		int one = ~0;   /* 0xff... */
		if(ioctl(fd, FIONBIO, &one) == -1)
			return -1;
	} else {
		int flags;

		flags = fcntl(fd, F_GETFL);
		if(flags == -1)
			return 0;	/* Okay if we can't get flags. */

		flags = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
		if(flags == -1)
			return -1;
	}

	return 0;
}


/*
 * Lock and unlock the file descriptor
 */

int
_xt_fd_lock(int fd, int poll_direction) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	if(xfd == NULL) return -1;

	if((poll_direction & POLLIN))
		_xt_mutex_lock(&xfd->fd_lock[0]);
	if((poll_direction & POLLOUT))
		_xt_mutex_lock(&xfd->fd_lock[1]);

	return 0;
}

int
_xt_fd_trylock(int fd, int poll_direction) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	if(xfd == NULL) return -1;

	if((poll_direction & POLLIN))
		if(_xt_mutex_trylock(&xfd->fd_lock[0]))
			return -1;
	if((poll_direction & POLLOUT))
		if(_xt_mutex_trylock(&xfd->fd_lock[1])) {
			if((poll_direction & POLLIN))
				_xt_mutex_unlock(&xfd->fd_lock[0]);
			return -1;
		}
	return 0;
}

int
_xt_fd_unlock(int fd, int poll_direction) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	if(xfd == NULL) return - 1;

	if((poll_direction & POLLIN))
		_xt_mutex_unlock(&xfd->fd_lock[0]);
	if((poll_direction & POLLOUT))
		_xt_mutex_unlock(&xfd->fd_lock[1]);

	return 0;
}


int
_xt_fd_wait(int fd, int poll_direction) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	int dir;

	if((xfd == NULL) || (xfd->fd_flags & XFD_INVALID)) {
		errno = EINVAL;
		return -1;
	}

	/* Clear the waiting status */
	if(poll_direction == 0) {
		if(xfd->waiter[0] == _xt_current_thread)
			xfd->waiter[0] = NULL;
		if(xfd->waiter[1] == _xt_current_thread)
			xfd->waiter[1] = NULL;
		return 0;
	}

	dir = (poll_direction & POLLIN) ? 0 : 1;

	/*
	 * Check if fd already has required condition.
	 */
	if((xfd->fd_flags & (dir ? XFD_WRITEABLE : XFD_READABLE)))
		return 0;

	/*
	 * If this descriptor was not placed into the queue, place it now.
	 */
	if((xfd->fd_flags & (dir ? XFD_KQUEUED_W : XFD_KQUEUED_R)) == 0) {
		struct kevent event;
		struct timespec ts = { 0, 0 };
		int ret;

		EV_SET(&event,
			fd,
			(dir == 0 ? EVFILT_READ : EVFILT_WRITE),
			EV_ADD,
			0,
			0,
			NULL
		);
		ret = kevent(_xt_kqueue, &event, 1, NULL, 0, &ts);
		if(ret == -1)
			return -1;

		/* Indicate we're monitoring it */
		xfd->fd_flags |= (dir ? XFD_KQUEUED_W : XFD_KQUEUED_R);
	}

	assert(xfd->waiter[dir] == NULL);

	xfd->waiter[dir] = _xt_current_thread;

	_xt_move2ring(_xt_current_thread, &_xt_sleeping_ring);

	return 0;
}


/*
 * Kernel queue staff
 */

int
_xt_is_queued(int fd, int interest) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	if(xfd == NULL) return -1;

	if((interest & POLLIN) && (xfd->fd_flags & XFD_KQUEUED_R))
		interest &= ~POLLIN;

	if((interest & POLLOUT) && (xfd->fd_flags & XFD_KQUEUED_W))
		interest &= ~POLLOUT;

	if(interest == 0) {
		return 1;	/* Every interest is present */
	} else {
		return 0;
	}

}

static void
__wakeup_waiter(xt_fd_t *xfd, int dir) {
	if(xfd->waiter[dir]) {

		_xt_move2ring(xfd->waiter[dir], &_xt_runnable_ring);

		/* Check opposite direction if it is the same */
		if(xfd->waiter[dir?0:1] == xfd->waiter[dir])
			xfd->waiter[dir?0:1] = NULL;

		xfd->waiter[dir] = NULL;
	}
}

int
_xt_gather_events(struct kevent *changelist, int nchanges, int tmo_ms) {
	xt_fd_t *xfd;
	struct timespec ts_s;
	struct timespec *ts = &ts_s;
	struct kevent *revent_list = keventlist_static;	/* Returned events */
	int revents;	/* Number of returned events */
	int events;
	int n;

	/*
	 * Construct timeout from milliseconds
	 */
	if(tmo_ms == 0) {
		ts->tv_sec = 0;
		ts->tv_nsec = 0;
	} else if(tmo_ms < 0) {
		ts = NULL;
	} else  {
		ts->tv_sec = tmo_ms / 1000;
		ts->tv_nsec = (tmo_ms % 1000) * 1000000;
	}

	/*
	 * Prepare returned event list.
	 * If the static keventlist is too small to hald possible
	 * all of changed events, point it to the changelist.
	 */
	revents = sizeof(keventlist_static) / sizeof(keventlist_static[0]);
	if(revents < nchanges) {
		revent_list = changelist;
		revents = nchanges;
	}

	events = kevent(_xt_kqueue, changelist, nchanges,
		revent_list, revents, ts);

	for(n = 0; n < events; n++) {

		xfd = _xt_get_xfd(revent_list[n].ident);
		if(xfd == NULL) {
			revent_list[n].flags = EV_DELETE;
			(void)kevent(_xt_kqueue, &revent_list[n], 1,
				NULL, 0, NULL);
			continue;
		}

		if((revent_list[n].flags & EV_ERROR)) {

			/* Set error state on fd */
			if(revent_list[n].fflags == EBADF) {
				xfd->fd_flags |= XFD_INVALID;
				xfd->fd_flags &=
					~(XFD_KQUEUED_R | XFD_KQUEUED_W);
			} else {
				xfd->fd_flags |= XFD_HUP;
			}

			/* Wake up waiting thread */
			__wakeup_waiter(xfd, 0);
			continue;
		} else {
			/* Clear error state on fd */
			xfd->fd_flags &= ~XFD_INVALID;
		}

		/*
		 * Processing incoming events
		 */

		switch(revent_list[n].filter) {
		case EVFILT_READ:
			xfd->fd_flags |= XFD_READABLE;

			/* Wake up waiting thread */
			__wakeup_waiter(xfd, 0);

			break;
		case EVFILT_WRITE:
			xfd->fd_flags |= XFD_WRITEABLE;

			/* Wake up waiting thread */
			__wakeup_waiter(xfd, 1);

			break;
		}

	}

	return 0;
}


/*
 * Return the poll()-like state of the descriptor.
 */
int
_xt_fd_state(int fd) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	int ret = 0;

	if(xfd && (xfd->fd_flags & XFD_INVALID) == 0) {
		if((xfd->fd_flags & XFD_READABLE))
			ret |= POLLIN;
		if((xfd->fd_flags & XFD_WRITEABLE))
			ret |= POLLOUT;
	} else {
		ret = POLLNVAL;
	}

	return ret;
}


void
_xt_fd_clear(int fd, enum xt_fd_flags fl) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	if(xfd)
		xfd->fd_flags &= ~fl;
}

void
_xt_xfd_close(int fd) {
	xt_fd_t *xfd = _xt_get_xfd(fd);
	if(xfd == NULL) return;

	xfd->fd_flags = XFD_INVALID;

	__wakeup_waiter(xfd, 0);
	__wakeup_waiter(xfd, 1);

	__xt_mutex_MM_unlock(&xfd->fd_lock[0]);
	__xt_mutex_MM_unlock(&xfd->fd_lock[1]);
}



/*
 * Helper functions
 */


static xt_fd_t *
_xt_get_xfd(int fd) {

	if(fd < 0) {
		errno = EBADF;
		return NULL;
	}

	if(fd >= xfd_size) {
		if(_xt_xfd_reinit())
			/* ENOMEM */
			return NULL;
		if((fd >= xfd_size)) {
			errno = EBADF;
			return NULL;
		}
	}

	return &xfd_wrapper[fd];
}

static int
_xt_xfd_reinit() {
	int size;

	size = getdtablesize();
	if(size > xfd_size) {
		int tmp_errno = errno;
		xt_fd_t *p;

		if(xfd_wrapper == fd_wrappers_static) {
			p = calloc(size, sizeof(xt_fd_t));
			if(p)
				memcpy(p, xfd_wrapper,
					xfd_size * sizeof(xt_fd_t));
		} else
			p = realloc(xfd_wrapper, size * sizeof(xt_fd_t));

		/* Restore errno */
		errno = tmp_errno;

		if(p == NULL) {
			return -1;
		} else  {
			memset(&p[xfd_size], 0,
				(size - xfd_size) * sizeof(xt_fd_t));
		}

		xfd_wrapper = p;
		xfd_size = size;
	}

	return 0;
}
