#include "xt_int.h"
#include "xt_timer.h"

/*
 * The original work on the data structures indirectly used in this
 * implementation was published by G. Varghese and A. Lauck in the paper
 * "Hashed and Hierarchical Timing Wheels: Data Structures for the Efficient
 * Implementation of a Timer Facility" in the Proceedings of the
 * 11th ACM Annual Symposium on Operating Systems Principles,
 * Austin, Texas Nov 1987.
 */


/*
 * Timer hash definitions.
 */
typedef struct xt_timer_hash_s {
	int numelements;
	int numbuckets;		/* 0 means "use internal" */

	xt_timer_t **buckets;
} xt_timer_hash_t;

/* Dynamic hashing support for timers */
static xt_timer_hash_t *__xt_timer_hash_new();
static void __xt_timer_hash_add(xt_timer_t *timer);
static void __xt_timer_hash_del(xt_timer_t *timer);

/* One statically defined timer hash */
static xt_timer_hash_t *timer_hash = NULL;


/*
 * Run-time persistent variables.
 */
struct timeval xt_current_time_tv;
static int _xt_current_tick;	/* Hardware (well, semi-) tick */
static int _xt_soft_tick;	/* Last tick we checked */
static int _xt_nearest_tick;	/* Tick number of nearest timer */



/*
 * Initialize internal timer stuff.
 */
int
_xt_timer_initialize() {

	timer_hash = __xt_timer_hash_new();
	if(timer_hash == NULL)
		return -1;

	/*
	 * Initialize start values of ticks.
 	 */
	gettimeofday(&xt_current_time_tv, NULL);
	_xt_current_tick = xt_current_time_tv.tv_sec * 100
		+ xt_current_time_tv.tv_usec / 10000;
	_xt_soft_tick = _xt_current_tick - 1;

	return 0;
}


/*
 * Set timer to wait specified number of heartbeats (hz)
 */
void
_xt_timer_start(xt_timer_t *opt_timer,
		int wait_ticks,
		void (*callback)(struct xt_timer_s *, void *data),
		void *data) {
	xt_timer_t *timer = opt_timer;

	if(timer == NULL)
		/* Not implemented */
		abort();

	/* Stop timer if timer was running */
	_xt_timer_stop(timer);

	/* Take a chance to test timers */
	_xt_timer_work();

	timer->callback = callback;
	timer->data = data;
	timer->dst_tick = _xt_current_tick + wait_ticks;
	timer->state = XTS_RUNNING;

	/* Add timer to ther timers hash */
	__xt_timer_hash_add(timer);

	if((_xt_nearest_tick - timer->dst_tick) > 0)
		_xt_nearest_tick = timer->dst_tick;
}


/*
 * Stop running the timer.
 */
void
_xt_timer_stop(xt_timer_t *timer) {

	/* Don't stop stopped timer */
	if(timer->state == XTS_STOPPED)
		return;

	__xt_timer_hash_del(timer);

	timer->state = XTS_STOPPED;
}


/*
 * Execute timers wheel.
 */
int
_xt_timer_work() {
	static int inside = 0;
	xt_timer_t *timer, *next;
	int ret = 0;

	/* Simple lock */
	if(inside++)
		goto finish;

	/*
	 * Compute current tick.
	 * Assume time may go backwards eventually.
	 */
	gettimeofday(&xt_current_time_tv, NULL);
	ret = (xt_current_time_tv.tv_sec * 100)
		+ (xt_current_time_tv.tv_usec / 10000);
	if((ret - _xt_current_tick) > 0)
		_xt_current_tick = ret;
	ret = 0;

	/*
	 * Don't do unnecessary amount of work if nearest timer
	 * is not right here.
	 */
	if(_xt_nearest_tick) {
		if(_xt_nearest_tick > _xt_current_tick) {
			ret = (_xt_nearest_tick
				- _xt_current_tick);
			goto finish;
		} else {
			/* Jump right to the nearest tick */
			_xt_soft_tick = _xt_nearest_tick - 1;
			_xt_nearest_tick = 0;
		}
	}

	/*
	 * Check all timer cells since the previous invocation.
	 */
	while(_xt_soft_tick != _xt_current_tick) {
		_xt_soft_tick++;

		timer = timer_hash->buckets[(_xt_soft_tick & 0x7fffffff)
			% timer_hash->numbuckets];
		if(timer == NULL)
			/* Empty bucket */
			continue;

		do {
			next = timer->h_next;

			if(timer->dst_tick == _xt_soft_tick) {
				/* Delete timer from the hash */
				__xt_timer_hash_del(timer);
				timer->state = XTS_STOPPED;
				timer->callback(timer, timer->data);
			}

		} while((timer = next));
	}

	/*
	 * Search for the nearest timer to cache its position.
	 */
	if(1) {
		int tick;
		int ticks;

		ret = 1000;	/* 100 ms */
		ticks = _xt_current_tick + ret;
		tick = _xt_soft_tick;

		while(tick != ticks) {
			tick++;

			timer = timer_hash->buckets[(tick & 0x7fffffff)
				% timer_hash->numbuckets];

			for(; timer; timer = timer->h_next) {
				/* Not our timer */
				if(timer->dst_tick == tick) {
					_xt_nearest_tick = tick;
					ret = (_xt_nearest_tick
						- _xt_current_tick);
					goto finish;
				}
			}
		}

		_xt_nearest_tick = tick;
		goto finish;
	}

finish:
	inside--;
	return ret;
}

/*
 * Sleep for specified amount of ticks.
 */
unsigned int
_xt_sleep(unsigned int ticks) {

	if(ticks == 0) {
		THR_SWITCH_CONTEXT;
		return 0;
	}

	/* Don't forget to wake me up ;) */
	_xt_schedule_wakeup(ticks);

	THR_SWITCH_CONTEXT;

	if(_xt_current_thread->timer.state == XTS_RUNNING) {
		int unslept =
			_xt_current_thread->timer.dst_tick
				- _xt_current_tick;
		if(unslept < 0)
			unslept = 0;

		_xt_timer_stop(&_xt_current_thread->timer);

		return unslept;
	}

	return 0;
}


static void
__xt_wakeup_thread_func(xt_timer_t *timer, void *thread_p) {
	xt_thread_t *thread = thread_p;

	_xt_move2ring(thread, &_xt_runnable_ring);
}

void
_xt_schedule_wakeup(unsigned int ticks) {

	_xt_timer_start(&_xt_current_thread->timer,
		ticks,
		__xt_wakeup_thread_func,
		_xt_current_thread);

	_xt_move2ring(_xt_current_thread, &_xt_sleeping_ring);
}


/*
 * Dynamic hashing support for timers.
 */


#define	ABSOLUTE_MAXIMUM_HASH_BUCKETS_NUMBER	(1024 * 1024)	/* 1M entries, 4M RAM */

static void
__xt_timer_hash_del(xt_timer_t *timer) {

	/* Found it; remove from lists and destruct as specified */
	if (timer->h_prev) {
		if((timer->h_prev->h_next = timer->h_next))
			timer->h_next->h_prev = timer->h_prev;
		
	} else {
		if((timer_hash->buckets[(timer->dst_tick & 0x7fffffff)
			% timer_hash->numbuckets] = timer->h_next))
			timer->h_next->h_prev = NULL;
	}

	timer->h_prev = NULL;
	timer->h_next = NULL;

	timer_hash->numelements--;
}

static int
__xt_timer_hash_expand(xt_timer_hash_t *h) {
	int newbuckets_count;
	xt_timer_t **newbuckets;

	if(h->numbuckets)
		newbuckets_count = h->numbuckets << 2;
	else
		newbuckets_count = 64;

	/* Too big hash table */
	if(h->numbuckets >= ABSOLUTE_MAXIMUM_HASH_BUCKETS_NUMBER)
		return -1;

	newbuckets = malloc(newbuckets_count * sizeof(void *));
	if(newbuckets == NULL)
		return -1;
	
	memset(newbuckets, 0, newbuckets_count * sizeof(void *));

	if(h->numbuckets) {
		xt_timer_t *timer, *next_timer;
		int old_bucket;
		int bucket;

		/*
		 * Rehash timers from old h->buckets to newbuckets.
		 */
		for(old_bucket = 0; old_bucket <= h->numbuckets; old_bucket++) {
			timer = h->buckets[old_bucket];
			if(timer == NULL)
				/* No timers in this buckets */
				continue;

			for(next_timer = timer->h_next; timer;
			    timer = next_timer, next_timer
					= (timer?timer->h_next:NULL)) {
				bucket = (timer->dst_tick & 0x7fffffff)
					% newbuckets_count;

				timer->h_prev = NULL;
				if((timer->h_next = newbuckets[bucket]))
					newbuckets[bucket]->h_prev = timer;
				newbuckets[bucket] = timer;
			}
		}

		free(h->buckets);
	}

	h->buckets = newbuckets;
	h->numbuckets = newbuckets_count;

	return 0;
}

static xt_timer_hash_t *
__xt_timer_hash_new() {
	xt_timer_hash_t *h;

	h = (xt_timer_hash_t *)malloc(sizeof(xt_timer_hash_t));
	if(h == NULL)
		return NULL;

	memset(h, 0, sizeof(xt_timer_hash_t));

	if(__xt_timer_hash_expand(h) == -1) {
		free(h);
		return NULL;
	}

	return h;
}

static void
__xt_timer_hash_add(xt_timer_t *timer) {
	xt_timer_t **bucket;

	if((timer_hash->numelements / timer_hash->numbuckets) > 2)
		(void)__xt_timer_hash_expand(timer_hash);

	bucket = &timer_hash->buckets[(timer->dst_tick & 0x7fffffff)
		% timer_hash->numbuckets];

	timer->h_prev = NULL;
	if((timer->h_next = *bucket))
		(*bucket)->h_prev = timer;
	*bucket = timer;

	timer_hash->numelements++;
}

